const mongoose = require("mongoose");
mongoose.Promise = global.Promise;
const slug = require("slugs");

const storeSchema = new mongoose.Schema(
  {
    name: {
      type: String,
      trim: true,
      required: "Please enter a store name!",
    },
    slug: String,
    description: {
      type: String,
      trim: true,
    },
    tags: [String],
    created: {
      type: Date,
      default: Date.now,
    },
    location: {
      type: {
        type: String,
        default: "Point",
      },
      coordinates: [
        {
          type: Number,
          required: "You must supply coordinates!",
        },
      ],
      address: {
        type: String,
        required: "You must supply an address!",
      },
    },
    photo: String,
    author: {
      type: mongoose.Schema.ObjectId,
      ref: "User",
      required: "You must supply an author",
    },
  },
  {
    toJSON: { virtuals: true },
    toObject: { virtuals: true },
  }
);

// Define our indexes
storeSchema.index({
  name: "text",
  description: "text",
});

storeSchema.index({ location: "2dsphere" });

storeSchema.pre("save", async function(next) {
  // not using arrow func bcz need to use this keyword
  if (!this.isModified("name")) {
    next(); // skip it
    return; // stop this function from running
  }
  this.slug = slug(this.name);
  // find other stores that have a slug of wes, wes-1, wes-2
  const slugRegEx = new RegExp(`^(${this.slug})((-[0-9]*$)?)$`, "i");
  const storesWithSlug = await this.constructor.find({ slug: slugRegEx });
  if (storesWithSlug.length) {
    this.slug = `${this.slug}-${storesWithSlug.length + 1}`;
  }
  next();
  // TODO make more resilient so slugs are unique
});

// Our custom static method for Store Schema
storeSchema.statics.getTagsList = function() {
  // Using a mongoDB aggregator pipeline operator here: $unwind
  return this.aggregate([
    { $unwind: "$tags" },
    // Another pipeline operator to group by id and add a property count to each group
    { $group: { _id: "$tags", count: { $sum: 1 } } },
    { $sort: { count: -1 } },
    // So these operators are working 1 by 1 - unwind - group - sort
  ]);
};

storeSchema.statics.getTopStores = function() {
  return this.aggregate([
    // Lookup stores and populate their reviews
    // the from value - reviews is actually our Review model, Mongodb auto makes it lowercase and adds a s at end
    { $lookup: { from: "reviews", localField: "_id", foreignField: "store", as: "reviews" } },
    // filter for only items that have 2 or more reviews
    { $match: { "reviews.1": { $exists: true } } },
    // add the average reviews field
    {
      $addFields: {
        averageRating: { $avg: "$reviews.rating" },
      },
    },
    // sort it by our own new field, highest reviews first
    { $sort: { averageRating: -1 } },
    // limit to at most 10
    { $limit: 10 },
  ]);
};

// find reviews where the stores _id property === reviews store property
storeSchema.virtual("reviews", {
  ref: "Review", // what model to link?
  localField: "_id", // which field on the store?
  foreignField: "store", // which field on the review?
});

function autopopulate(next) {
  this.populate("reviews");
  next();
}

// pre-hook to run autopopulate whenever find or findOne is called
storeSchema.pre("find", autopopulate);
storeSchema.pre("findOne", autopopulate);

module.exports = mongoose.model("Store", storeSchema);
